import time
import logging


CALLBACK_RETRIES = 10


def log_exec_time(name):
    def decorator(func):
        def exec_time(*args, **kwargs):
            start = time.perf_counter()
            func(*args, **kwargs)
            stop = time.perf_counter()
            logging.debug(f"Executed {name} in {(stop - start):.04f}s")
        return exec_time
    return decorator


def retry_callback(func):
    def execute(*args, **kwargs):
        for index in range(CALLBACK_RETRIES):
            try:
                func(*args, **kwargs)
                return
            except Exception as e:
                if index == CALLBACK_RETRIES - 1:
                    logging.error(f"Callback failure={e}")
                else:
                    time.sleep(0.01)
    return execute



from ctypes import (
    Structure,
    c_bool,
    c_float,
    c_int,
    c_int64,
    c_wchar,
    wintypes,
)

MAX_STRING_SIZE = 1024
SHORT_STRING_SIZE = 256

MAX_ACTION_INFO = 256
MAX_ACTIONS_TO_DO = 32

MAX_RELATION_TARGETS = 25
MAX_RELATIONS = 5

MAX_HYPERLINKS = 64

MAX_TABLE_SELECTIONS = 64

MAX_KEY_BINDINGS = 10


class JavaObject(c_int64):
    pass


class AccessibleContextInfo(Structure):
    _fields_ = [
        ('name', c_wchar * MAX_STRING_SIZE),
        ('description', c_wchar * MAX_STRING_SIZE),
        ('role', c_wchar * SHORT_STRING_SIZE),
        ('role_en_US', c_wchar * SHORT_STRING_SIZE),
        ('states', c_wchar * SHORT_STRING_SIZE),
        ('states_en_US', c_wchar * SHORT_STRING_SIZE),
        ('indexInParent', c_int),
        ('childrenCount', c_int),
        ('x', c_int),
        ('y', c_int),
        ('width', c_int),
        ('height', c_int),
        ('accessibleComponent', wintypes.BOOL),
        ('accessibleAction', wintypes.BOOL),
        ('accessibleSelection', wintypes.BOOL),
        ('accessibleText', wintypes.BOOL),
        ('accessibleValue', wintypes.BOOL)
    ]


class AccessBridgeVersionInfo(Structure):
    _fields_ = [
        ('VMversion', c_wchar * SHORT_STRING_SIZE),
        ('bridgeJavaClassVersion', c_wchar * SHORT_STRING_SIZE),
        ('bridgeJavaDLLVersion', c_wchar * SHORT_STRING_SIZE),
        ('bridgeWinDLLVersion', c_wchar * SHORT_STRING_SIZE)
    ]


class AccessibleTextInfo(Structure):
    _fields_ = [
        ('charCount', c_int),
        ('caretIndex', c_int),
        ('indexAtPoint', c_int)
    ]


class AccessibleTextItemsInfo(Structure):
    _fields_ = [
        ('letter', c_wchar),
        ('word', c_wchar * SHORT_STRING_SIZE),
        ('sentence', c_wchar * MAX_STRING_SIZE)
    ]


class AccessibleTextSelectionInfo(Structure):
    _fields_ = [
        ('selectionStartIndex', c_int),
        ('selectionEndIndex', c_int),
        ('selectedText', c_wchar * MAX_STRING_SIZE)
    ]


class AccessibleTextAttributesInfo(Structure):
    _fields_ = [
        ("bold", wintypes.BOOL),
        ("italic", wintypes.BOOL),
        ("underline", wintypes.BOOL),
        ("strikethrough", wintypes.BOOL),
        ("superscript", wintypes.BOOL),
        ("subscript", wintypes.BOOL),

        ("backgroundColor", c_wchar * SHORT_STRING_SIZE),
        ("foregroundColor", c_wchar * SHORT_STRING_SIZE),
        ("fontFamily", c_wchar * SHORT_STRING_SIZE),
        ("fontSize", c_int),

        ("alignment", c_int),
        ("bidiLevel", c_int),

        ("firstLineIndent", c_float),
        ("leftIndent", c_float),
        ("rightIndent", c_float),
        ("lineSpacing", c_float),
        ("spaceAbove", c_float),
        ("spaceBelow", c_float),

        ("fullAttributesString", c_wchar * MAX_STRING_SIZE),
    ]


class AccessibleTextRectInfo(Structure):
    _fields_ = [
        ("x", c_int),
        ("y", c_int),
        ("width", c_int),
        ("height", c_int)
    ]


class AccessibleActionInfo(Structure):
    _fields_ = [
        ("name", c_wchar * SHORT_STRING_SIZE)
    ]


class AccessibleActions(Structure):
    _fields_ = [
        ("actionsCount", c_int),
        ("actionInfo", AccessibleActionInfo * MAX_ACTION_INFO)
    ]


class AccessibleActionsToDo(Structure):
    _fields_ = [
        ("actionsCount", c_int),
        ("actions", AccessibleActionInfo * MAX_ACTION_INFO)
    ]


class AccessibleRelationInfo(Structure):
    _fields_ = [
        ("key", c_wchar * SHORT_STRING_SIZE),
        ("targetCount", c_int),
        ("targets", JavaObject * MAX_RELATION_TARGETS)
    ]


class AccessibleRelationSetInfo(Structure):
    _fields = [
        ("relationCount", c_int),
        ("AccessibleRelationInfo", AccessibleRelationInfo * MAX_RELATIONS)
    ]


class AccessibleHyperlinkInfo(Structure):
    _fields_ = [
        ("text", c_wchar * SHORT_STRING_SIZE),
        ("startIndex", c_int),
        ("endIndex", c_int),
        ("accessibleHyperlink", JavaObject)
    ]


class AccessibleHypertextInfo(Structure):
    _fields_ = [
        ("linkCount", c_int),
        ("links", AccessibleHyperlinkInfo * MAX_HYPERLINKS),
        ("accessibleHypertext", JavaObject)
    ]


class AccessibleKeyBindingInfo(Structure):
    _fields_ = [
        ("character", c_wchar),
        ("modifiers", c_int)
    ]


class AccessibleKeyBindings(Structure):
    _fields_ = [
        ("keyBindingsCount", c_int),
        ("AccessibleKeyBindingInfo", AccessibleKeyBindingInfo * MAX_KEY_BINDINGS)
    ]


class AccessibleTableInfo(Structure):
    _fields_ = [
        ("caption", JavaObject),
        ("summary", JavaObject),
        ("rowCount", c_int),
        ("columnCount", c_int),
        ("accessibleContext", JavaObject),
        ("accessibleTable", JavaObject)
    ]


class AccessibleTableCellInfo(Structure):
    _fields = [
        ("accessibleContext", JavaObject),
        ("index", c_int),
        ("row", c_int),
        ("column", c_int),
        ("rowExtent", c_int),
        ("columnExtent", c_int),
        ("isSelected", c_bool)
    ]



import logging
import threading
from typing import Dict, List

from JABWrapper.jab_wrapper import JavaAccessBridgeWrapper
from JABWrapper.jab_types import (
    AccessibleActionsToDo,
    AccessibleHypertextInfo,
    AccessibleKeyBindings,
    AccessibleTableInfo,
    AccessibleTextAttributesInfo,
    AccessibleTextItemsInfo,
    AccessibleTextRectInfo,
    JavaObject,
    AccessibleContextInfo,
    AccessibleTextInfo,
    AccessibleTextSelectionInfo,
    MAX_HYPERLINKS
)
from JABWrapper.utils import log_exec_time, retry_callback


class SearchElement:
    def __init__(self, name, value) -> None:
        self.name = name
        self.value = value


class _Parser:
    def parse() -> None:
        raise NotImplementedError()


class _AccessibleTableParser(_Parser):
    def __init__(self, aci: AccessibleContextInfo) -> None:
        self._aci = aci
        self.table = AccessibleTableInfo()

    def __str__(self) -> str:
        if self._aci.role == "table":
            return f" table={self.table.rowCount},{self.table.columnCount}"
        return ""

    def parse(self, jab_wrapper: JavaAccessBridgeWrapper, context: JavaObject) -> None:
        if self._aci.role == "table":
            self.table = jab_wrapper.get_accessible_table_info(context)


class _AccessibleKeyBindingsParser(_Parser):
    def __init__(self, aci: AccessibleContextInfo) -> None:
        self._aci = aci
        self.key_bindings = AccessibleKeyBindings()

    def __str__(self) -> str:
        string = f" kbs={self.key_bindings.keyBindingsCount}"
        if self.key_bindings.keyBindingsCount > 0:
            for index in range(self.key_bindings.keyBindingsCount):
                info = self.key_bindings.AccessibleKeyBindingInfo[index]
                string += f" c={info.character} m={info.modifiers}"
        return string

    def parse(self, jab_wrapper: JavaAccessBridgeWrapper, context: JavaObject) -> None:
        self.key_bindings = jab_wrapper.get_accessible_key_bindings(context)


class _AccessibleActionsParser(_Parser):
    def __init__(self, aci: AccessibleContextInfo) -> None:
        self._aci = aci
        self.actions = dict()

    def __str__(self) -> str:
        if not self._aci.accessibleAction:
            return ""
        return " actions={}".format(", ".join([action for action in self.actions]))

    def parse(self, jab_wrapper: JavaAccessBridgeWrapper, context: JavaObject) -> None:
        self.actions.clear()
        if self._aci.accessibleAction:
            actions = jab_wrapper.get_accessible_actions(context)
            for index in range(actions.actionsCount):
                actionInfo = actions.actionInfo[index]
                self.actions[actionInfo.name.lower()] = actionInfo

    def do_action(self, jab_wrapper: JavaAccessBridgeWrapper, context: JavaObject, action: str) -> None:
        if not action.lower() in self.actions:
            raise NotImplementedError("Does not implement the {} action".format(action))
        actions = AccessibleActionsToDo(actionsCount=1, actions=(self.actions[action],))
        jab_wrapper.do_accessible_actions(context, actions)

    def click(self, jab_wrapper: JavaAccessBridgeWrapper, context: JavaObject) -> None:
        self.do_action(jab_wrapper, context, "click")

    def insert_content(self, jab_wrapper: JavaAccessBridgeWrapper, context: JavaObject, text: str) -> None:
        jab_wrapper.set_text_contents(context, text)


class _AccessibleHypertextParser(_Parser):
    def __init__(self, aci: AccessibleContextInfo) -> None:
        self._aci = aci
        self.info = AccessibleHypertextInfo()

    def __str__(self) -> str:
        if not self._aci.accessibleText or self.info.linkCount > MAX_HYPERLINKS or self.info.linkCount < 0:
            return ""
        txt = f" links={self.info.linkCount}"
        for i in range(self.info.linkCount):
            txt += f", link={self.info.links[i].text}"
        return txt

    def parse(self, jab_wrapper: JavaAccessBridgeWrapper, context: JavaObject) -> None:
        """
        TODO: Identify which elements may contain hypertext.

        From java documentation (https://docs.oracle.com/javase/6/docs/api/javax/accessibility/AccessibleHypertext.html),
        the AccessibleText element may implement the AccessibleHypertext object, but not all AccessibleText elements do.

        The element JEditorPane.JEditorPaneAccessibleHypertextSupport does implement it, but is identifiable from the AccessibleContextInfo?
        """
        if self._aci.accessibleText:
            self.info = jab_wrapper.get_accessible_hypertext(context)


class _AccessibleValueParser(_Parser):
    def __init__(self, aci: AccessibleContextInfo) -> None:
        self._aci = aci
        self.value = u''
        self.min = u''
        self.max = u''

    def parse(self, jab_wrapper: JavaAccessBridgeWrapper, context: JavaObject) -> None:
        if self._aci.accessibleValue:
            self.value = jab_wrapper.get_current_accessible_value_from_context(context)
            self.min = jab_wrapper.get_minimum_accessible_value_from_context(context)
            self.max = jab_wrapper.get_maximum_accessible_value_from_context(context)

    def __str__(self) -> str:
        if not self._aci.accessibleValue:
            return ""
        return " v={};".format(self.value)


class _AccessibleTextParser(_Parser):
    def __init__(self, aci: AccessibleContextInfo) -> None:
        self._aci = aci
        self.info = AccessibleTextInfo()
        self.items = AccessibleTextItemsInfo()
        self.selection = AccessibleTextSelectionInfo()
        self.attributes_info = AccessibleTextAttributesInfo()
        self.rect_info = AccessibleTextRectInfo()

    def parse(self, jab_wrapper: JavaAccessBridgeWrapper, context: JavaObject) -> None:
        if self._aci.accessibleText:
            self.info = jab_wrapper.get_context_text_info(context, self._aci.x, self._aci.y)
            self.items = jab_wrapper.get_accessible_text_items(context, 0)
            self.selection = jab_wrapper.get_accessible_text_selection_info(context)
            self.attributes_info = jab_wrapper.get_accessible_text_attributes(context, 0)
            self.rect_info = jab_wrapper.get_accessible_text_rect(context, 0)

    def __str__(self) -> str:
        if not self._aci.accessibleText:
            return ""
        return " tc={} w={} s={} st={};".format(
            self.info.charCount,
            self.items.word,
            self.items.sentence,
            self.selection.selectedText
        )


class ContextNode:
    def __init__(self, jab_wrapper: JavaAccessBridgeWrapper, context: JavaObject, lock, ancestry: int = 0) -> None:
        self._jab_wrapper = jab_wrapper
        self._lock = lock
        self.ancestry = ancestry
        self.context: JavaObject = context
        self.state = None
        self.virtual_accessible_name = None
        self._parse_context()
        self.children: list[ContextNode] = []
        self._parse_children()

    def _parse_context(self) -> None:
        logging.debug(f"Parsing element={self.context}")
        self.aci: AccessibleContextInfo = self._jab_wrapper.get_context_info(self.context)
        logging.debug(f"Parsed element role={self.aci.role}, name={self.aci.name}")
        self.virtual_accessible_name = self._jab_wrapper.get_virtual_accessible_name(self.context)
        self.atp = _AccessibleTextParser(self.aci)
        self.avp = _AccessibleValueParser(self.aci)
        self._aap = _AccessibleActionsParser(self.aci)
        self.akbp = _AccessibleKeyBindingsParser(self.aci)
        self.htp = _AccessibleHypertextParser(self.aci)
        self.tp = _AccessibleTableParser(self.aci)
        self._parsers: List[_Parser] = [self.atp, self.avp, self._aap, self.akbp, self.htp, self.tp]
        [parser.parse(self._jab_wrapper, self.context) for parser in self._parsers]

    def _parse_children(self) -> None:
        for i in range(0, self.aci.childrenCount):
            child_context = self._jab_wrapper.get_child_context(self.context, i)
            child_node = ContextNode(self._jab_wrapper, child_context, self._lock, self.ancestry + 1)
            self.children.append(child_node)

    def __repr__(self) -> str:
        """
        Returns a string that represents the object tree with detailed Node values
        """
        string = "{}C={}, Role={}, Name={}, VAN={}, Desc={}, St={}, Sts={}, at x={}:y={} w={} h={}; cc={};".format(
            '  ' * self.ancestry,
            self.context,
            repr(self.aci.role),
            repr(self.aci.name),
            repr(self.virtual_accessible_name),
            repr(self.aci.description),
            repr(self.state),
            repr(self.aci.states),
            self.aci.x,
            self.aci.y,
            self.aci.width,
            self.aci.height,
            self.aci.childrenCount
        )
        for parser in self._parsers:
            string += f"{parser}"
        for child in self.children:
            string += f"\n{repr(child)}"
        return string

    def __str__(self) -> str:
        """
        Returns a string of Node values
        """
        string = "Role={}, Name={}, VAN={}, Desc={}, St={}, Sts={}, at x={}:y={} w={} h={}; cc={};".format(
            repr(self.aci.role),
            repr(self.aci.name),
            repr(self.virtual_accessible_name),
            repr(self.aci.description),
            repr(self.state),
            repr(self.aci.states),
            self.aci.x,
            self.aci.y,
            self.aci.width,
            self.aci.height,
            self.aci.childrenCount
        )
        for parser in self._parsers:
            string += "{}".format(parser)
        return string

    def _get_node_by_context(self, context: JavaObject):
        if self._jab_wrapper.is_same_object(self.context, context):
            return self
        else:
            for child in self.children:
                node = child._get_node_by_context(context)
                if node:
                    return node

    def _update_node(self) -> None:
        self.children = []
        self._parse_context()
        self._parse_children()

    def _van_match(self, search_elements: List[SearchElement]) -> bool:
        for search_element in search_elements:
            if search_element.name == "VAN":
                if search_element.value != self.virtual_accessible_name:
                    return False
        return True

    def get_by_attrs(self, search_elements: List[SearchElement]) -> List:
        """
        Get element with given seach attributes.

        The SearchElement object takes a name of the field and the field value:
        element = context_tree.get_by_attrs([SearchElement("role", "text")])

        Returns an array of matching elements.
        """
        with self._lock:
            elements = list()
            van_match = self._van_match(search_elements)
            found = all([getattr(self.aci, search_element.name).startswith(search_element.value) for search_element in search_elements])
            if found and van_match:
                elements.append(self)
            for child in self.children:
                child_elements = child.get_by_attrs(search_elements)
                elements.extend(child_elements)
            return elements

    def request_focus(self) -> None:
        """
        Request focus for element
        """
        with self._lock:
            self._jab_wrapper.request_focus(self.context)

    def get_actions(self) -> Dict:
        """
        Get all actions available for element
        """
        with self._lock:
            return self._aap.actions

    def do_action(self, action: str) -> None:
        """
        Do any action found with the get_actions interface

        Will raise APIException is action is not available
        """
        with self._lock:
            self._aap.do_action(self._jab_wrapper, self.context, action)

    def click(self) -> None:
        """
        Do click action for element
        """
        with self._lock:
            self._aap.click(self._jab_wrapper, self.context)

    def insert_text(self, text: str) -> None:
        """
        Do insert content action for element
        """
        with self._lock:
            self._aap.insert_content(self._jab_wrapper, self.context, text)


class ContextTree:
    @log_exec_time("Init context tree")
    def __init__(self, jab_wrapper: JavaAccessBridgeWrapper) -> None:
        self._lock = threading.RLock()
        self._jab_wrapper = jab_wrapper
        self.root = ContextNode(jab_wrapper, jab_wrapper.context, self._lock)
        self._register_callbacks()

    def __repr__(self) -> str:
        return f"{repr(self.root)}"

    def __str__(self):
        return f"{self.root}"

    @retry_callback
    def _property_change_cp(self, source: JavaObject, property: str, old_value: str, new_value: str) -> None:
        with self._lock:
            node: ContextNode = self.root._get_node_by_context(source)
            if node:
                setattr(node.aci, property, new_value)
                logging.debug(f"Property={property} changed from={old_value} to={new_value} for node={node}")

    @retry_callback
    def _property_name_change_cp(self, source: JavaObject, old_value: str, new_value: str) -> None:
        with self._lock:
            node: ContextNode = self.root._get_node_by_context(source)
            if node:
                setattr(node.aci, "name", new_value)
                logging.debug(f"Name changed from={old_value} to={new_value} for node={node}")

    @retry_callback
    def _property_description_change_cp(self, source: JavaObject, old_value: str, new_value: str) -> None:
        with self._lock:
            node: ContextNode = self.root._get_node_by_context(source)
            if node:
                setattr(node.aci, "description", new_value)
                logging.debug(f"Description changed from={old_value} to={new_value} for node={node}")

    @retry_callback
    def _property_state_change_cp(self, source: JavaObject, old_value: str, new_value: str) -> None:
        with self._lock:
            node: ContextNode = self.root._get_node_by_context(source)
            if node:
                node.state = new_value
                logging.debug(f"State changed from={old_value} to={new_value} for node={node}")

    @retry_callback
    def _property_value_change_cp(self, source: JavaObject, old_value: str, new_value: str) -> None:
        with self._lock:
            node: ContextNode = self.root._get_node_by_context(source)
            if node:
                node.avp.value = new_value
                logging.debug(f"Value changed from={old_value} to={new_value} for node={node}")

    @retry_callback
    def _property_selection_change_cp(self, source: JavaObject) -> None:
        with self._lock:
            node: ContextNode = self.root._get_node_by_context(source)
            if node:
                node._parse_context()
                logging.debug(f"Selected text changed for node={node}")

    @retry_callback
    def _property_text_change_cp(self, source: JavaObject) -> None:
        with self._lock:
            node: ContextNode = self.root._get_node_by_context(source)
            if node:
                node._parse_context()
                logging.debug(f"Text changed for node={node}")

    @retry_callback
    def _property_caret_change_cp(self, source: JavaObject, old_pos: int, new_pos: int) -> None:
        # Caret information is not stored in the node component
        logging.debug("Property caret change event ignored")

    @retry_callback
    def _visible_data_change_cp(self, source: JavaObject) -> None:
        with self._lock:
            node = self.root._get_node_by_context(source)
            if node:
                node._update_node()
                logging.debug(f"Visible data changed for node tree={repr(node)}")

    @retry_callback
    def _property_child_change_cp(self, source: JavaObject, old_child: JavaObject, new_child: JavaObject) -> None:
        # Not needed to track as the visibility change event handles the coordinate update
        logging.debug("Property child change event ignored")

    @retry_callback
    def _property_active_descendent_change_cp(self, source: JavaObject, old_child: JavaObject, new_child: JavaObject) -> None:
        # The activity status is not stored inside the tree model
        logging.debug("Property active descendent change event ignored")

    @retry_callback
    def _property_table_model_change_cp(self, source: JavaObject, old_value: str, new_value: str) -> None:
        # TODO: Add table model parsing
        logging.debug("Property table model change event ignored")

    @retry_callback
    def _menu_selected_cp(self, source: JavaObject) -> None:
        # All menu events can be ignored as the visibility change event already gives needed information update to the context tree
        logging.debug("Menu selected event ignored")

    @retry_callback
    def _menu_deselected_cp(self, source: JavaObject) -> None:
        # All menu events can be ignored as the visibility change event already gives needed information update to the context tree
        logging.debug("Menu deselected event ignored")

    @retry_callback
    def _menu_canceled_cp(self, source: JavaObject) -> None:
        # All menu events can be ignored as the visibility change event already gives needed information update to the context tree
        logging.debug("Menu canceled event ignored")

    @retry_callback
    def _focus_gained_cp(self, source: JavaObject) -> None:
        # State information is not stored in the context tree
        logging.debug("Focus gained event ignored")

    @retry_callback
    def _focus_lost_cp(self, source: JavaObject) -> None:
        # State information is not stored in the context tree
        logging.debug("Focus lost event ignored")

    @retry_callback
    def _caret_update_cp(self, source: JavaObject) -> None:
        # Caret information is not stored in the context tree
        logging.debug("Caret update event ignored")

    @retry_callback
    def _mouse_clicked_cp(self, source: JavaObject) -> None:
        # Ignore the mouse events, as the change events will update the context tree
        logging.debug("mouse clicked event ignored")

    @retry_callback
    def _mouse_entered_cp(self, source: JavaObject) -> None:
        # Ignore the mouse events, as the change events will update the context tree
        logging.debug("mouse entered event ignored")

    @retry_callback
    def _mouse_exited_cp(self, source: JavaObject) -> None:
        # Ignore the mouse events, as the change events will update the context tree
        logging.debug("mouse exited event ignored")

    @retry_callback
    def _mouse_pressed_cp(self, source: JavaObject) -> None:
        # Ignore the mouse events, as the change events will update the context tree
        logging.debug("mouse pressed event ignored")

    @retry_callback
    def _mouse_released_cp(self, source: JavaObject) -> None:
        # Ignore the mouse events, as the change events will update the context tree
        logging.debug("mouse released event ignored")

    @retry_callback
    def _popup_menu_canceled_cp(self, source: JavaObject) -> None:
        # Ignore the popup events
        logging.debug("popup menu canceled event ignored")

    @retry_callback
    def _popup_menu_will_become_invisible_cp(self, source: JavaObject) -> None:
        # Ignore the popup events
        logging.debug("popup menu will become invisible event ignored")

    @retry_callback
    def _popup_menu_will_become_visible_cp(self, source: JavaObject) -> None:
        # Ignore the popup events
        logging.debug("popup menu will become visible event ignored")

    def _register_callbacks(self) -> None:
        """
        Register callbacks to the jab wrapper when context updated events
        are generated from the Access Bridge
        """
        # Property change event handlers
        self._jab_wrapper.register_callback("property_change", self._property_change_cp)
        self._jab_wrapper.register_callback("property_name_change", self._property_name_change_cp)
        self._jab_wrapper.register_callback("property_description_change", self._property_description_change_cp)
        self._jab_wrapper.register_callback("property_state_change", self._property_state_change_cp)
        self._jab_wrapper.register_callback("property_value_change", self._property_value_change_cp)
        self._jab_wrapper.register_callback("property_selection_change", self._property_selection_change_cp)
        self._jab_wrapper.register_callback("property_text_change", self._property_text_change_cp)
        self._jab_wrapper.register_callback("property_caret_change", self._property_caret_change_cp)
        self._jab_wrapper.register_callback("property_visible_data_change", self._visible_data_change_cp)
        self._jab_wrapper.register_callback("property_child_change", self._property_child_change_cp)
        self._jab_wrapper.register_callback("property_active_descendent_change", self._property_active_descendent_change_cp)
        self._jab_wrapper.register_callback("property_table_model_change", self._property_table_model_change_cp)
        # Menu event handlers
        self._jab_wrapper.register_callback("menu_selected", self._menu_selected_cp)
        self._jab_wrapper.register_callback("menu_deselected", self._menu_deselected_cp)
        self._jab_wrapper.register_callback("menu_calceled", self._menu_canceled_cp)
        # Focus event handlers
        self._jab_wrapper.register_callback("focus_gained", self._focus_gained_cp)
        self._jab_wrapper.register_callback("focus_lost", self._focus_lost_cp)
        # Caret update event handler
        self._jab_wrapper.register_callback("caret_update", self._caret_update_cp)
        # Mouse events
        self._jab_wrapper.register_callback("mouse_clicked", self._mouse_clicked_cp)
        self._jab_wrapper.register_callback("mouse_entered", self._mouse_entered_cp)
        self._jab_wrapper.register_callback("mouse_exited", self._mouse_exited_cp)
        self._jab_wrapper.register_callback("mouse_pressed", self._mouse_pressed_cp)
        self._jab_wrapper.register_callback("mouse_released", self._mouse_released_cp)
        # Popup menu events
        self._jab_wrapper.register_callback("popup_menu_canceled", self._popup_menu_canceled_cp)
        self._jab_wrapper.register_callback("popup_menu_will_become_invisible", self._popup_menu_will_become_invisible_cp)
        self._jab_wrapper.register_callback("popup_menu_will_become_visible", self._popup_menu_will_become_visible_cp)

    def get_by_attrs(self, search_elements: List[SearchElement]) -> List[ContextNode]:
        """
        Find an element from the context tree.

        The root node has the same API as each child node inside the tree.

        The SearchElement object takes a name of the field and the field value:
        element = context_tree.get_by_attrs([SearchElement("role", "text")])

        Returns an array of matching elements.
        """
        return self.root.get_by_attrs(search_elements)



import os
import sys
import time
import logging

from ctypes import (
    CFUNCTYPE,
    c_long,
    c_short,
    c_void_p,
    c_wchar,
    c_wchar_p,
    wintypes,
    POINTER,
    byref,
    windll,
    c_int,
    cdll,
    WINFUNCTYPE,
    WinError,
    create_unicode_buffer
)

from typing import Callable, List, Tuple

from JABWrapper.jab_types import (
    AccessBridgeVersionInfo,
    AccessibleActions,
    AccessibleActionsToDo,
    AccessibleHyperlinkInfo,
    AccessibleHypertextInfo,
    AccessibleKeyBindings,
    AccessibleRelationSetInfo,
    AccessibleTableCellInfo,
    AccessibleTextAttributesInfo,
    AccessibleTableInfo,
    AccessibleTextItemsInfo,
    AccessibleTextRectInfo,
    JavaObject,
    AccessibleContextInfo,
    AccessibleTextInfo,
    AccessibleTextSelectionInfo,
    MAX_STRING_SIZE,
    SHORT_STRING_SIZE
)


logging_file_handler = logging.FileHandler("jab_wrapper.log", "w", "utf-8")
logging_file_handler.setFormatter(logging.Formatter("%(asctime)s [%(threadName)s] [%(levelname)s] %(message)s"))
logging_file_handler.setLevel(logging.DEBUG)

logging_stream_handler = logging.StreamHandler(sys.stdout)
logging_stream_handler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s] %(message)s"))
logging_stream_handler.setLevel(logging.INFO)

logging.basicConfig(
    level=logging.DEBUG,
    handlers=[logging_file_handler, logging_stream_handler]
)


# https://stackoverflow.com/questions/21175922/enumerating-windows-trough-ctypes-in-python
WNDENUMPROC = WINFUNCTYPE(wintypes.BOOL, wintypes.HWND, wintypes.LPARAM)
user32 = windll.user32
user32.EnumWindows.argtypes = [
    WNDENUMPROC,
    wintypes.LPARAM]
user32.GetWindowTextLengthW.argtypes = [
    wintypes.HWND]
user32.GetWindowTextW.argtypes = [
    wintypes.HWND,
    wintypes.LPWSTR,
    c_int]


PropertyChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject, c_wchar_p, c_wchar_p, c_wchar_p)
PropertyNameChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject, c_wchar_p, c_wchar_p)
PropertyDescriptionChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject, c_wchar_p, c_wchar_p)
PropertStateChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject, c_wchar_p, c_wchar_p)
PropertyValueChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject, c_wchar_p, c_wchar_p)
PropertySelectionChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
PropertyTextChangedFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
PropertyCaretChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject, c_int, c_int)
PropertyVisibleDataChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
PropertyChildChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject, JavaObject, JavaObject)
PropertyActiveDescendentChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject, JavaObject, JavaObject)
PropertyTableModelChangeFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject, c_wchar_p, c_wchar_p)
FocusGainedFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
FocusLostFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
CaretUpdateFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
MouseClickedFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
MouseEnteredFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
MouseExitedFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
MousePressedFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
MouseReleasedFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
MenuSelectedFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
MenuDeselectedFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
MenuCanceledFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
PopupMenuCanceledFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
PopupMenuWillBecomeInvisibleFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)
PopupMenuWillBecomeVisibleFP = CFUNCTYPE(None, c_long, JavaObject, JavaObject)


class APIException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class _ReleaseEvent:
    def __init__(self, context, vmID, name, event, source) -> None:
        self._context = context
        self._vmID = vmID
        self._name = name
        self._event = event
        self._source = source
        self._start_exec: float = 0

    def __enter__(self):
        logging.debug(f"Received {self._name} event={self._source}")
        self._start_exec = time.perf_counter()

    def __exit__(self, type, value, traceback):
        stop_exec = time.perf_counter()
        logging.debug(f"Executed {self._name} in {(stop_exec - self._start_exec):.04f}s")
        self._context._wab.releaseJavaObject(self._vmID, self._event)


class JavaAccessBridgeWrapper:
    def __init__(self) -> None:
        logging.debug("Loading WindowsAccessBridge")
        if "RC_JAVA_ACCESS_BRIDGE_DLL" not in os.environ:
            raise OSError("Environment variable: RC_JAVA_ACCESS_BRIDGE_DLL not found")
        if not os.path.isfile(os.path.normpath(os.environ['RC_JAVA_ACCESS_BRIDGE_DLL'])):
            raise FileNotFoundError(f"File not found: {os.environ['RC_JAVA_ACCESS_BRIDGE_DLL']}")
        self._wab: cdll = cdll.LoadLibrary(os.path.normpath(os.environ['RC_JAVA_ACCESS_BRIDGE_DLL']))
        logging.debug("WindowsAccessBridge loaded succesfully")
        self._define_functions()
        self._define_callbacks()
        self._set_callbacks()
        self._wab.Windows_run()

        self._hwnd: wintypes.HWND = None
        self._vmID = c_long()
        self.context = JavaObject()

        self._current_window: str = ''

        # Any reader can register callbacks here that are executed when AccessBridge events are seen
        self._context_callbacks: dict[str, List[Callable[[JavaObject], None]]] = dict()

    def shutdown(self):
        self._context_callbacks = dict()
        self._remove_callbacks()

    def _define_functions(self) -> None:
        # void Windows_run()
        self._wab.Windows_run.argtypes = []
        self._wab.Windows_run.restype = None

        # void ReleaseJavaObject(long vmID, AccessibleContext context)
        self._wab.releaseJavaObject.argtypes = [c_long, JavaObject]
        self._wab.releaseJavaObject.restype = None

        # BOOL GetVersionInfo(long vmID, AccessBridgeVersionInfo *info)
        self._wab.getVersionInfo.argtypes = [c_long, POINTER(AccessBridgeVersionInfo)]
        self._wab.getVersionInfo.restype = wintypes.BOOL

        # Accessible context
        # BOOL isJavaWindow(HWND window)
        self._wab.isJavaWindow.argtypes = [wintypes.HWND]
        self._wab.isJavaWindow.restype = wintypes.BOOL
        # BOOL isSameObject(long vmID, AccessibleContext context_from, AccessibleContext context_to)
        self._wab.isSameObject.argtypes = [c_long, JavaObject, JavaObject]
        self._wab.isSameObject.restypes = wintypes.BOOL
        # BOOL GetAccessibleContextFromHWND(HWND window, long *vmID, AccessibleContext *context)
        self._wab.getAccessibleContextFromHWND.argtypes = [wintypes.HWND, POINTER(c_long), POINTER(JavaObject)]
        self._wab.getAccessibleContextFromHWND.restype = wintypes.BOOL
        # HWND getHWNDFromAccessibleContext(long vmID, AccessibleContext context)
        self._wab.getHWNDFromAccessibleContext.argtypes = [c_long, JavaObject]
        self._wab.getHWNDFromAccessibleContext.restype = wintypes.HWND
        # BOOL getAccessibleContextAt(long vmID, AccessibleContext parent, int x, int y, AccessibleContext *context)
        self._wab.getAccessibleContextAt.argtypes = [c_long, JavaObject, c_int, c_int, POINTER(JavaObject)]
        self._wab.getAccessibleContextAt.restype = wintypes.BOOL
        # TODO: getAccessibleContextWithFocus
        # BOOL getAccessibleContextInfo(long vmID, AccessibleContext context, AccessibleContextInfo *info)
        self._wab.getAccessibleContextInfo.argtypes = [c_long, JavaObject, POINTER(AccessibleContextInfo)]
        self._wab.getAccessibleContextInfo.restype = wintypes.BOOL
        # AccessibleContext getAccessibleChildFromContext(long vmID, AccessibleContext context, int integer)
        self._wab.getAccessibleChildFromContext.argtypes = [c_long, JavaObject, c_int]
        self._wab.getAccessibleChildFromContext.restype = JavaObject
        # JavaObject getAccessibleParentFromContext(c_long vmID, JavaObject child_context)
        self._wab.getAccessibleParentFromContext.argtypes = [c_long, JavaObject]
        self._wab.getAccessibleParentFromContext.restype = JavaObject

        # Accessible table
        # BOOL getAccessibleTableInfo(long vmID, AccessibleContext context, AccessibleTableInfo *tableInfo)
        self._wab.getAccessibleTableInfo.argtypes = [c_long, JavaObject, POINTER(AccessibleTableInfo)]
        self._wab.getAccessibleTableInfo.restype = wintypes.BOOL
        # BOOL getAccessibleTableCellInfo(long vmID, AccessibleTable accessibleTable, int row, int column, AccessibleTableCellInfo *tableCellInfo)
        self._wab.getAccessibleTableCellInfo.argtypes = [c_long, JavaObject, c_int, c_int, POINTER(AccessibleTableCellInfo)]
        self._wab.getAccessibleTableCellInfo.restype = wintypes.BOOL
        # BOOL getAccessibleTableRowHeader(long vmID, AccessibleContext context, AccessibleTableInfo *tableInfo)
        self._wab.getAccessibleTableRowHeader.argtypes = [c_long, JavaObject, POINTER(AccessibleTableInfo)]
        self._wab.getAccessibleTableRowHeader.restype = wintypes.BOOL
        # BOOL getAccessibleTableColumnHeader(long vmID, AccessibleContext context, AccessibleTableInfo *tableInfo)
        self._wab.getAccessibleTableColumnHeader.argtypes = [c_long, JavaObject, POINTER(AccessibleTableInfo)]
        self._wab.getAccessibleTableColumnHeader.restype = wintypes.BOOL
        # JavaObject getAccessibleTableRowDescription(long vmID, AccessibleContext context, int row)
        self._wab.getAccessibleTableRowDescription.argtypes = [c_long, JavaObject, c_int]
        self._wab.getAccessibleTableRowDescription.restype = JavaObject
        # JavaObject getAccessibleTableColumnDescription(long vmID, AccessibleContext context, int row)
        self._wab.getAccessibleTableColumnDescription.argtypes = [c_long, JavaObject, c_int]
        self._wab.getAccessibleTableColumnDescription.restype = JavaObject
        # int getAccessibleTableRowSelectionCount(long vmID, AccessibleTable table)
        self._wab.getAccessibleTableRowSelectionCount.argtypes = [c_long, JavaObject]
        self._wab.getAccessibleTableRowSelectionCount.restype = c_int
        # BOOL isAccessibleTableRowSelected(long vmID, AccessibleTable table, int row)
        self._wab.isAccessibleTableRowSelected.argtypes = [c_long, JavaObject, c_int]
        self._wab.isAccessibleTableRowSelected.restype = wintypes.BOOL
        # TODO: What interface is this? The java API doesn't shed any more light than the windows bridge implementation
        # int getAccessibleTableRowSelections(long vmID, AccessibleTable table, int count, int *selections)
        # int getAccessibleTableColumnSelectionCount(long vmID, AccessibleTable table)
        self._wab.getAccessibleTableColumnSelectionCount.argtypes = [c_long, JavaObject]
        self._wab.getAccessibleTableColumnSelectionCount.restype = c_int
        # BOOL isAccessibleTableColumnSelected(long vmID, AccessibleTable table, int row)
        self._wab.isAccessibleTableColumnSelected.argtypes = [c_long, JavaObject, c_int]
        self._wab.isAccessibleTableColumnSelected.restype = wintypes.BOOL
        # TODO: What interface is this? The java API doesn't shed any more light than the windows bridge implementation
        # int getAccessibleTableColumnSelections(long vmID, AccessibleTable table, int count, int *selections)
        # int getAccessibleTableRow(long vmID, AccessibleTable table, int index)
        self._wab.getAccessibleTableRow.argtypes = [c_long, JavaObject, c_int]
        self._wab.getAccessibleTableRow.restype = c_int
        # int getAccessibleTableColumn(long vmID, AccessibleTable table, int index)
        self._wab.getAccessibleTableColumn.argtypes = [c_long, JavaObject, c_int]
        self._wab.getAccessibleTableColumn.restype = c_int
        # int getAccessibleTableIndex(long vmID, AccessibleTable table, int row)
        self._wab.getAccessibleTableIndex.argtypes = [c_long, JavaObject, c_int, c_int]
        self._wab.getAccessibleTableIndex.restype = c_int

        # AccessibleRelationSet
        # BOOL getAccessibleRelationSet(long vmID, AccessibleContext accessibleContext, AccessibleRelationSetInfo *relationSetInfo)
        self._wab.getAccessibleRelationSet.argtypes = [c_long, JavaObject, POINTER(AccessibleRelationSetInfo)]
        self._wab.getAccessibleRelationSet.restype = wintypes.BOOL

        # AccessibleHypertext
        # BOOL getAccessibleHypertext(long vmID, AccessibleContext accessibleContext, AccessibleHypertextInfo *hypertextInfo)
        self._wab.getAccessibleHypertext.argtypes = [c_long, JavaObject, POINTER(AccessibleHypertextInfo)]
        self._wab.getAccessibleHypertext.restype = wintypes.BOOL
        # BOOL activateAccessibleHyperlink(long vmID, AccessibleContext accessibleContext, AccessibleHyperlink accessibleHyperlink)
        self._wab.activateAccessibleHyperlink.argtypes = [c_long, JavaObject, JavaObject]
        self._wab.activateAccessibleHyperlink.restype = wintypes.BOOL
        # BOOL getAccessibleHyperlinkCount(long vmID, AccessibleContext accessibleContext)
        self._wab.getAccessibleHyperlinkCount.argtypes = [c_long, JavaObject]
        self._wab.getAccessibleHyperlinkCount.restype = c_int
        # BOOL getAccessibleHypertextExt(long vmID, AccessibleContext accessibleContext, int nStartIndex, AccessibleHypertextInfo *hypertextInfo)
        self._wab.getAccessibleHypertextExt.argtypes = [c_long, JavaObject, c_int, POINTER(AccessibleHypertextInfo)]
        self._wab.getAccessibleHypertextExt.restype = wintypes.BOOL
        # BOOL getAccessibleHypertextLinkIndex(long vmID, AccessibleContext accessibleContext, int charIndex, int *linkIndex)
        self._wab.getAccessibleHypertextLinkIndex.argtypes = [c_long, JavaObject, c_int]
        self._wab.getAccessibleHypertextLinkIndex.restype = c_int
        # BOOL getAccessibleHyperlink(long vmID, AccessibleContext accessibleContext, int index, AccessibleHyperlinkInfo *hyperlinkInfo)
        self._wab.getAccessibleHyperlink.argtypes = [c_long, JavaObject, c_int, POINTER(AccessibleHyperlinkInfo)]
        self._wab.getAccessibleHyperlink.restype = wintypes.BOOL

        # Accessible KeyBindings, Icons and Actions
        # BOOL getAccessibleKeyBindings(long vmID, AccessibleContext context, AccessibleKeyBindings *bindings)
        self._wab.getAccessibleKeyBindings.argtypes = [c_long, JavaObject, POINTER(AccessibleKeyBindings)]
        self._wab.getAccessibleKeyBindings.restypes = wintypes.BOOL
        # TODO: getAccessibleIcons
        # BOOL getAccessibleActions(long vmID, AccessibleContext context, AccessibleActions *actions)
        self._wab.getAccessibleActions.argtypes = [c_long, JavaObject, POINTER(AccessibleActions)]
        self._wab.getAccessibleActions.restypes = wintypes.BOOL
        # BOOL doAccessibleActions(long vmID, AccessibleContext context, AccessibleActionsToDo actionsToDo, bool *result, int *failure_index)
        self._wab.doAccessibleActions.argtypes = [c_long, JavaObject, AccessibleActionsToDo, POINTER(c_int)]
        self._wab.doAccessibleActions.restypes = wintypes.BOOL

        # AccessibleText
        # AccessibleTextInfo GetAccessibleTextInfo(long vmID, AccessibleContext context, AccessibleTextInfo *info, int x, int y)
        self._wab.getAccessibleTextInfo.argtypes = [c_long, JavaObject, POINTER(AccessibleTextInfo), c_int, c_int]
        self._wab.getAccessibleTextInfo.restype = wintypes.BOOL
        # AccesibleTextItems GetAccessibleTextItems(long vmID, AccessibleContext context, AccesibleTextItems *items, int index)
        self._wab.getAccessibleTextItems.argtypes = [c_long, JavaObject, POINTER(AccessibleTextItemsInfo), c_int]
        self._wab.getAccessibleTextItems.restype = wintypes.BOOL
        # BOOL GetAccessibleTextSelectionInfo(long vmID, AccessibleContext context, AccessibleTextSelectionInfo *textSelection)
        self._wab.getAccessibleTextSelectionInfo.argtypes = [c_long, JavaObject, POINTER(AccessibleTextSelectionInfo)]
        self._wab.getAccessibleTextSelectionInfo.restype = wintypes.BOOL
        # BOOL getAccessibleTextAttributes(long vmID, AccessibleContext context, int index, AccessibleTextAttributesInfo *attributesInfo)
        self._wab.getAccessibleTextAttributes.argtypes = [c_long, JavaObject, c_int, POINTER(AccessibleTextAttributesInfo)]
        self._wab.getAccessibleTextAttributes.restype = wintypes.BOOL
        # BOOL getAccessibleTextRect(long vmID, AccessibleContext context, AccessibleTextRectInfo *rectInfo, int index)
        self._wab.getAccessibleTextRect.argtypes = [c_long, JavaObject, POINTER(AccessibleTextRectInfo), c_int]
        self._wab.getAccessibleTextRect.restype = wintypes.BOOL
        # BOOL getAccessibleTextLineBounds(long vmID, AccessibleContext context, int index, int *startIndex, int *endIndex)
        self._wab.getAccessibleTextLineBounds.argtypes = [c_long, JavaObject, c_int, POINTER(c_int), POINTER(c_int)]
        self._wab.getAccessibleTextLineBounds.restype = wintypes.BOOL
        # BOOL getAccessibleTextRange(long vmID, AccessibleContext context, int start, int end, c_wchar_p *text, short len)
        self._wab.getAccessibleTextRange.argtypes = [c_long, JavaObject, c_int, c_int, c_wchar_p, c_short]
        self._wab.getAccessibleTextRange.restype = wintypes.BOOL

        # AccessibleValue
        # BOOL getCurrentAccessibleValueFromContext(long vmID, AccessibleContext context, c_wchar *value, short len)
        self._wab.getCurrentAccessibleValueFromContext.argtypes = [c_long, JavaObject, c_wchar_p, c_short]
        self._wab.getCurrentAccessibleValueFromContext.restype = wintypes.BOOL
        # BOOL getMaximumAccessibleValueFromContext(long vmID, AccessibleContext context, c_wchar *value, short len)
        self._wab.getMaximumAccessibleValueFromContext.argtypes = [c_long, JavaObject, c_wchar_p, c_short]
        self._wab.getMaximumAccessibleValueFromContext.restype = wintypes.BOOL
        # BOOL getMinimumAccessibleValueFromContext(long vmID, AccessibleContext context, c_wchar *value, short len)
        self._wab.getMinimumAccessibleValueFromContext.argtypes = [c_long, JavaObject, c_wchar_p, c_short]
        self._wab.getMinimumAccessibleValueFromContext.restype = wintypes.BOOL

        # AccessibleSelection
        # TODO: addAccessibleSelectionFromContext
        # TODO: clearAccessibleSelectionFromContext
        # TODO: getAccessibleSelectionFromContext
        # TODO: getAccessibleSelectionCountFromContext
        # TODO: isAccessibleChildSelectedFromContext
        # TODO: removeAccessibleSelectionFromContext
        # TODO: selectAllAccessibleSelectionFromContext

        # Utility
        # BOOL setTextContents(long vmID, AccessibleContext context, str text)
        self._wab.setTextContents.argtypes = [c_long, JavaObject, c_wchar * MAX_STRING_SIZE]
        self._wab.setTextContents.restypes = wintypes.BOOL
        # TODO: getParentWithRole
        # TODO: getParentWithRoleElseRoot
        # TODO: getTopLevelObject
        # TODO: getObjectDepth
        # TODO: getActiveDescendent
        # BOOL getVirtualAccessibleNameFP(long vmID, AccessibleContext context, str name, int len)
        self._wab.getVirtualAccessibleName.argtypes = [c_long, JavaObject, c_wchar * MAX_STRING_SIZE, c_int]
        self._wab.getVirtualAccessibleName.restype = wintypes.BOOL
        # BOOL requestFocus(long vmID, AccessibleContext context)
        self._wab.requestFocus.argtypes = [c_long, JavaObject]
        self._wab.requestFocus.restypes = wintypes.BOOL
        # TODO: selectTextRange
        # TODO: getTextAttributesInRange
        # TODO: getVisibleChildrenCount
        # TODO: getVisibleChildren
        # TODO: setCaretPosition
        # TODO: getCaretLocation
        # TODO: getEventsWaitingFP

    def _define_callbacks(self) -> None:
        # Property events
        self._wab.setPropertyChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyChangeFP.restype = None
        self._wab.setPropertyNameChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyNameChangeFP.restype = None
        self._wab.setPropertyDescriptionChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyDescriptionChangeFP.restype = None
        self._wab.setPropertyStateChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyStateChangeFP.restype = None
        self._wab.setPropertyValueChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyValueChangeFP.restype = None
        self._wab.setPropertySelectionChangeFP.argtypes = [c_void_p]
        self._wab.setPropertySelectionChangeFP.restype = None
        self._wab.setPropertyTextChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyTextChangeFP.restype = None
        self._wab.setPropertyCaretChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyCaretChangeFP.restype = None
        self._wab.setPropertyVisibleDataChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyVisibleDataChangeFP.restype = None
        self._wab.setPropertyChildChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyChildChangeFP.restype = None
        self._wab.setPropertyActiveDescendentChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyActiveDescendentChangeFP.restype = None
        self._wab.setPropertyTableModelChangeFP.argtypes = [c_void_p]
        self._wab.setPropertyTableModelChangeFP.restype = None
        # Menu events
        self._wab.setMenuSelectedFP.argtypes = [c_void_p]
        self._wab.setMenuSelectedFP.restype = None
        self._wab.setMenuDeselectedFP.argtypes = [c_void_p]
        self._wab.setMenuDeselectedFP.restype = None
        self._wab.setMenuCanceledFP.argtypes = [c_void_p]
        self._wab.setMenuCanceledFP.restype = None
        # Focus events
        self._wab.setFocusGainedFP.argtypes = [c_void_p]
        self._wab.setFocusGainedFP.restype = None
        self._wab.setFocusLostFP.argtypes = [c_void_p]
        self._wab.setFocusLostFP.restype = None
        # Caret update events
        self._wab.setCaretUpdateFP.argtypes = [c_void_p]
        self._wab.setCaretUpdateFP.restype = None
        # Mouse events
        self._wab.setMouseClickedFP.argtypes = [c_void_p]
        self._wab.setMouseClickedFP.restype = None
        self._wab.setMouseEnteredFP.argtypes = [c_void_p]
        self._wab.setMouseEnteredFP.restype = None
        self._wab.setMouseExitedFP.argtypes = [c_void_p]
        self._wab.setMouseExitedFP.restype = None
        self._wab.setMousePressedFP.argtypes = [c_void_p]
        self._wab.setMousePressedFP.restype = None
        self._wab.setMouseReleasedFP.argtypes = [c_void_p]
        self._wab.setMouseReleasedFP.restype = None
        # Popup menu events
        self._wab.setPopupMenuCanceledFP.argtypes = [c_void_p]
        self._wab.setPopupMenuCanceledFP.restype = None
        self._wab.setPopupMenuWillBecomeInvisibleFP.argtypes = [c_void_p]
        self._wab.setPopupMenuWillBecomeInvisibleFP.restype = None
        self._wab.setPopupMenuWillBecomeVisibleFP.argtypes = [c_void_p]
        self._wab.setPopupMenuWillBecomeVisibleFP.restype = None

    def _set_callbacks(self) -> None:
        # Property events
        self._wab.setPropertyChangeFP(self._get_callback_func("setPropertyChangeFP", PropertyChangeFP, self.property_change))
        self._wab.setPropertyNameChangeFP(self._get_callback_func("setPropertyNameChangeFP", PropertyNameChangeFP, self.property_name_change))
        self._wab.setPropertyDescriptionChangeFP(self._get_callback_func("setPropertyDescriptionChangeFP", PropertyDescriptionChangeFP,
                                                                         self.property_description_change))
        self._wab.setPropertyStateChangeFP(self._get_callback_func("setPropertyStateChangeFP", PropertStateChangeFP, self.property_state_change))
        self._wab.setPropertyValueChangeFP(self._get_callback_func("setPropertyValueChangeFP", PropertyValueChangeFP, self.property_value_change))
        self._wab.setPropertySelectionChangeFP(self._get_callback_func("setPropertySelectionChangeFP", PropertySelectionChangeFP,
                                                                       self.property_selection_change))
        self._wab.setPropertyTextChangeFP(self._get_callback_func("setPropertyTextChangeFP", PropertyTextChangedFP, self.property_text_change))
        self._wab.setPropertyCaretChangeFP(self._get_callback_func("setPropertyCaretChangeFP", PropertyCaretChangeFP, self.property_caret_change))
        self._wab.setPropertyVisibleDataChangeFP(self._get_callback_func("setPropertyVisibleDataChangeFP", PropertyVisibleDataChangeFP,
                                                                         self.property_visible_data_change))
        self._wab.setPropertyChildChangeFP(self._get_callback_func("setPropertyChildChangeFP", PropertyChildChangeFP, self.property_child_change))
        self._wab.setPropertyActiveDescendentChangeFP(self._get_callback_func("setPropertyActiveDescendentChangeFP", PropertyActiveDescendentChangeFP,
                                                                              self.property_active_descendent_change))
        self._wab.setPropertyTableModelChangeFP(self._get_callback_func("setPropertyTableModelChangeFP", PropertyTableModelChangeFP,
                                                                        self.property_table_model_change))
        # Menu events
        self._wab.setMenuSelectedFP(self._get_callback_func("setMenuSelectedFP", MenuSelectedFP, self.menu_selected))
        self._wab.setMenuDeselectedFP(self._get_callback_func("setMenuDeselectedFP", MenuDeselectedFP, self.menu_deselected))
        self._wab.setMenuCanceledFP(self._get_callback_func("setMenuCanceledFP", MenuCanceledFP, self.menu_canceled))
        # Focus events
        self._wab.setFocusGainedFP(self._get_callback_func("setFocusGainedFP", FocusGainedFP, self.focus_gained))
        self._wab.setFocusLostFP(self._get_callback_func("setFocusLostFP", FocusLostFP, self.focus_lost))
        # Caret update events
        self._wab.setCaretUpdateFP(self._get_callback_func("setCaretUpdateFP", CaretUpdateFP, self.caret_update))
        # Mouse events
        self._wab.setMouseClickedFP(self._get_callback_func("setMouseClickedFP", MouseClickedFP, self.mouse_clicked))
        self._wab.setMouseEnteredFP(self._get_callback_func("SetMouseEnteredFP", MouseEnteredFP, self.mouse_entered))
        self._wab.setMouseExitedFP(self._get_callback_func("setMouseExitedFP", MouseExitedFP, self.mouse_exited))
        self._wab.setMousePressedFP(self._get_callback_func("setMousePressedFP", MousePressedFP, self.mouse_pressed))
        self._wab.setMouseReleasedFP(self._get_callback_func("setMouseReleasedFP", MouseReleasedFP, self.mouse_released))
        # Popup menu events
        self._wab.setPopupMenuCanceledFP(self._get_callback_func("setPopupMenuCanceledFP", PopupMenuCanceledFP, self.popup_menu_canceled))
        self._wab.setPopupMenuWillBecomeInvisibleFP(self._get_callback_func("setPopupMenuWillBecomeInvisibleFP", PopupMenuWillBecomeInvisibleFP,
                                                    self.popup_menu_will_become_invisible))
        self._wab.setPopupMenuWillBecomeVisibleFP(self._get_callback_func("setPopupMenuWillBecomeVisibleFP", PopupMenuWillBecomeVisibleFP,
                                                  self.popup_menu_will_become_visible))

    def _remove_callbacks(self) -> None:
        # Property events
        self._wab.setPropertyChangeFP(None)
        self._wab.setPropertyNameChangeFP(None)
        self._wab.setPropertyDescriptionChangeFP(None)
        self._wab.setPropertyStateChangeFP(None)
        self._wab.setPropertyValueChangeFP(None)
        self._wab.setPropertySelectionChangeFP(None)
        self._wab.setPropertyTextChangeFP(None)
        self._wab.setPropertyCaretChangeFP(None)
        self._wab.setPropertyVisibleDataChangeFP(None)
        self._wab.setPropertyChildChangeFP(None)
        self._wab.setPropertyActiveDescendentChangeFP(None)
        self._wab.setPropertyTableModelChangeFP(None)
        # Menu events
        self._wab.setMenuSelectedFP(None)
        self._wab.setMenuDeselectedFP(None)
        self._wab.setMenuCanceledFP(None)
        # Focus events
        self._wab.setFocusGainedFP(None)
        self._wab.setFocusLostFP(None)
        # Caret update events
        self._wab.setCaretUpdateFP(None)
        # Mouse events
        self._wab.setMouseClickedFP(None)
        self._wab.setMouseEnteredFP(None)
        self._wab.setMouseExitedFP(None)
        self._wab.setMousePressedFP(None)
        self._wab.setMouseReleasedFP(None)
        # Popup menu events
        self._wab.setPopupMenuCanceledFP(None)
        self._wab.setPopupMenuWillBecomeInvisibleFP(None)
        self._wab.setPopupMenuWillBecomeVisibleFP(None)

    def _enumerate_windows(self, hwnd, lParam) -> bool:
        if not hwnd:
            logging.error("Invalid window handle={hwnd}")
            return True

        length = user32.GetWindowTextLengthW(hwnd) + 1
        buffer = create_unicode_buffer(length)
        user32.GetWindowTextW(hwnd, buffer, length)
        isJava = False
        title = buffer.value
        try:
            isJava = self._wab.isJavaWindow(hwnd)
        except OSError as e:
            logging.error(f"Failed to enumerate window={hwnd} error={e}")
            return True
        if isJava:
            logging.debug(f"found java window={title}")
            if self._current_window == title:
                logging.debug(f"found matching window={title}")
                self._hwnd = hwnd
                self._vmID = c_long()
                self.context = JavaObject()
                self._wab.getAccessibleContextFromHWND(self._hwnd, byref(self._vmID), byref(self.context))

        return True

    def set_hwnd(self, hwnd: wintypes.HWND) -> None:
        self._hwnd = hwnd

    def set_context(self, vm_id: c_long, context: JavaObject) -> None:
        self._vmID = vm_id
        self.context = context

    def get_current_windows_handle(self) -> wintypes.HWND:
        return self._hwnd

    def release_object(self, context: JavaObject) -> None:
        if self._vmID and self.context:
            logging.debug(f"Releasing object={context}")
            self._wab.releaseJavaObject(self._vmID, c_long(context.value).value)

    def switch_window_by_title(self, title: str) -> None:
        """
        Switch the context to window by title
        """
        # Add the title as the current context and find the correct window
        self._current_window = title
        self._vmID = None
        self.context = None
        windows = WNDENUMPROC(self._enumerate_windows)
        if not windll.user32.EnumWindows(windows, 0):
            raise WinError()
        if not self._hwnd or not self._vmID or not self.context:
            raise Exception("Window not found")
        logging.info("Found Java window text={}, hwnd={} vmID={} context={}\n".format(
            self._current_window,
            self._hwnd,
            self._vmID,
            self.context,
        ))

    def get_accessible_context_from_hwnd(self, hwnd: wintypes.HWND) -> Tuple[c_long, JavaObject]:
        vm_id = c_long()
        context = JavaObject()
        ok = self._wab.getAccessibleContextFromHWND(hwnd, byref(vm_id), byref(context))
        if not ok:
            raise APIException("Failed to get accessible context from HWND")
        return vm_id, context

    def get_hwnd_from_accessible_context(self, context) -> wintypes.HWND:
        return self._wab.getHWNDFromAccessibleContext(self._vmID, context)

    def get_accessible_context_at(self, parent: JavaObject, x: int, y: int) -> JavaObject:
        context = JavaObject()
        ok = self._wab.getAccessibleContextAt(self._vmID, parent, x, y, byref(context))
        if not ok:
            raise APIException("Failed to get accessible context at={x},{y}")
        return context

    def get_child_context(self, context: JavaObject, index: int, ) -> JavaObject:
        return self._wab.getAccessibleChildFromContext(self._vmID, context, index)

    def get_accessible_parent_from_context(self, context) -> JavaObject:
        return self._wab.getAccessibleParentFromContext(self._vmID, context)

    def get_accessible_table_info(self, context: JavaObject) -> AccessibleTableInfo:
        table_info = AccessibleTableInfo()
        ok = self._wab.getAccessibleTableInfo(self._vmID, context, byref(table_info))
        if not ok:
            raise APIException("Failed to get accessible table info")
        return table_info

    def get_accessible_table_cell_info(self, table_context: JavaObject, row: int, column: int) -> AccessibleTableCellInfo:
        table_cell_info = AccessibleTableCellInfo()
        ok = self._wab.getAccessibleTableCellInfo(self._vmID, table_context, row, column, byref(table_cell_info))
        if not ok:
            raise APIException(f"Failed to get accessible table cell info at={row},{column}")
        return table_cell_info

    def get_accessible_table_row_header(self, context) -> AccessibleTableInfo:
        table_info = AccessibleTableInfo()
        ok = self._wab.getAccessibleTableRowHeader(self._vmID, context, byref(table_info))
        if not ok:
            raise APIException("Failed to get accessible table row header")
        return table_info

    def get_accessible_table_column_header(self, context) -> AccessibleTableInfo:
        table_info = AccessibleTableInfo()
        ok = self._wab.getAccessibleTableColumnHeader(self._vmID, context, byref(table_info))
        if not ok:
            raise APIException("Failed to get accessible table column header")
        return table_info

    def get_accessible_table_row_description(self, context: JavaObject, row: int) -> JavaObject:
        return self._wab.getAccessibleTableRowDescription(self._vmID, context, row)

    def get_accessible_table_column_description(self, context: JavaObject, column: int) -> JavaObject:
        return self._wab.getAccessibleTableColumnDescription(self._vmID, context, column)

    def get_accessible_table_row_selection_count(self, table_context: JavaObject) -> int:
        return self._wab.getAccessibleTableRowSelectionCount(self._vmID, table_context)

    def is_accessible_table_row_selected(self, table_context: JavaObject, row: int) -> bool:
        return self._wab.isAccessibleTableRowSelected(self._vmID, table_context, row)

    def get_accessible_table_row(self, table_context: JavaObject, index: int) -> int:
        return self._wab.getAccessibleTableRow(self._vmID, table_context, index)

    def get_accessible_table_column(self, table_context: JavaObject, index: int) -> int:
        return self._wab.getAccessibleTableColumn(self._vmID, table_context, index)

    def get_accessible_table_index(self, table_context: JavaObject, row: int, column: int) -> int:
        return self._wab.getAccessibleTableIndex(self._vmID, table_context, row, column)

    def get_accessible_table_column_selection_count(self, table_context: JavaObject) -> int:
        return self._wab.getAccessibleTableColumnSelectionCount(self._vmID, table_context)

    def is_accessible_table_column_selected(self, table_context: JavaObject, row: int) -> bool:
        return self._wab.isAccessibleTableColumnSelected(self._vmID, table_context, row)

    def get_accessible_relation_set_info(self, context: JavaObject) -> AccessibleRelationSetInfo:
        relation_set_info = AccessibleRelationSetInfo()
        logging.info("getting rel set")
        ok = self._wab.getAccessibleRelationSet(self._vmID, context, byref(relation_set_info))
        logging.info(f"rel set={ok}")
        if not ok:
            raise APIException("Failed to get accessible relation set info")
        return relation_set_info

    def get_context_info(self, context: JavaObject) -> AccessibleContextInfo:
        info = AccessibleContextInfo()
        ok = self._wab.getAccessibleContextInfo(self._vmID, context, byref(info))
        if not ok:
            raise APIException("Failed to get accessible context info")
        return info

    def get_version_info(self) -> AccessBridgeVersionInfo:
        info = AccessBridgeVersionInfo()
        ok = self._wab.getVersionInfo(self._vmID, byref(info))
        if not ok:
            raise APIException("Failed to get version info")
        return info

    def get_accessible_hypertext(self, context: JavaObject) -> AccessibleHypertextInfo:
        hypertext_info = AccessibleHypertextInfo()
        ok = self._wab.getAccessibleHypertext(self._vmID, context, byref(hypertext_info))
        if not ok:
            raise APIException("Failed to get accessible hypertext info")
        return hypertext_info

    def activate_accessible_hyperlink(self, context, hyperlink: JavaObject) -> None:
        ok = self._wab.activateAccessibleHyperlink(self._vmID, context, hyperlink)
        if not ok:
            raise APIException("Failed to activate accessible hypertext link")

    def get_accessible_hyperlink_count(self, context: JavaObject) -> int:
        return self._wab.getAccessibleHyperlinkCount(self._vmID, context)

    def get_accessible_hypertext_ext(self, context: JavaObject, index: int) -> AccessibleHypertextInfo:
        hypertext_info = AccessibleHypertextInfo()
        ok = self._wab.getAccessibleHypertextExt(self._vmID, context, index, byref(hypertext_info))
        if not ok:
            raise APIException(f"Failed to get accessible hypertext info starting from index={index}")
        return hypertext_info

    def get_accessible_hypertext_link_index(self, hypertext: JavaObject, char_index: int) -> int:
        return self._wab.getAccessibleHypertextLinkIndex(self._vmID, hypertext, char_index)

    def get_accessible_hyperlink(self, hypertext: JavaObject, index: int) -> AccessibleHyperlinkInfo:
        hyperlink_info = AccessibleHyperlinkInfo()
        ok = self._wab.getAccessibleHyperlink(self._vmID, hypertext, index, byref(hyperlink_info))
        if not ok:
            raise APIException(f"Failed to get accessible hypertext link info at index={index}")
        return hyperlink_info

    def get_context_text_info(self, context: JavaObject, x: int, y: int) -> AccessibleTextInfo:
        info = AccessibleTextInfo()
        ok = self._wab.getAccessibleTextInfo(self._vmID, context, byref(info), x, y)
        if not ok:
            raise APIException("Failed to get accessible text info")
        return info

    def get_accessible_text_items(self, context: JavaObject, index: int) -> AccessibleTextItemsInfo:
        info = AccessibleTextItemsInfo()
        ok = self._wab.getAccessibleTextItems(self._vmID, context, byref(info), index)
        if not ok:
            raise APIException("Failed to get accessible text  context")
        return info

    def get_accessible_text_selection_info(self, context: JavaObject) -> AccessibleTextSelectionInfo:
        info = AccessibleTextSelectionInfo()
        ok = self._wab.getAccessibleTextSelectionInfo(self._vmID, context, byref(info))
        if not ok:
            raise APIException("Failed to get accessible text selection info")
        return info

    def get_accessible_text_attributes(self, context: JavaObject, index: int) -> AccessibleTextAttributesInfo:
        attributes_info = AccessibleTextAttributesInfo()
        ok = self._wab.getAccessibleTextAttributes(self._vmID, context, index, attributes_info)
        if not ok:
            raise APIException("Failed to get accessible text attributes info")
        return attributes_info

    def get_accessible_text_rect(self, context: JavaObject, index: int) -> AccessibleTextRectInfo:
        rect_info = AccessibleTextRectInfo()
        ok = self._wab.getAccessibleTextRect(self._vmID, context, byref(rect_info), index)
        if not ok:
            raise APIException("Failed to get accessible text rect info")
        return rect_info

    def get_accessible_text_line_bounds(self, context, index) -> Tuple[int, int]:
        start_index = c_int()
        end_index = c_int()
        ok = self._wab.getAccessibleTextLineBounds(self._vmID, context, index, byref(start_index), byref(end_index))
        if not ok:
            raise APIException(f"Failed to get accessible text line bounds at={index}")
        return start_index, end_index

    def get_accessible_text_range(self, context: JavaObject, start_index: int, end_index: int, length: c_short) -> str:
        buf = create_unicode_buffer(length)
        ok = self._wab.getAccessibleTextRange(self._vmID, context, start_index, end_index, buf, length)
        if not ok:
            raise APIException("Failed to get accessible range")
        return buf.value

    def get_current_accessible_value_from_context(self, context: JavaObject) -> str:
        buf = create_unicode_buffer(SHORT_STRING_SIZE)
        ok = self._wab.getCurrentAccessibleValueFromContext(self._vmID, context, buf, SHORT_STRING_SIZE)
        if not ok:
            raise APIException("Failed to get current accessible value from context")
        return buf.value

    def get_maximum_accessible_value_from_context(self, context: JavaObject) -> str:
        buf = create_unicode_buffer(SHORT_STRING_SIZE)
        ok = self._wab.getMaximumAccessibleValueFromContext(self._vmID, context, buf, SHORT_STRING_SIZE)
        if not ok:
            raise APIException("Failed to get maximum accessible value from context")
        return buf.value

    def get_minimum_accessible_value_from_context(self, context: JavaObject) -> str:
        buf = create_unicode_buffer(SHORT_STRING_SIZE)
        ok = self._wab.getMinimumAccessibleValueFromContext(self._vmID, context, buf, SHORT_STRING_SIZE)
        if not ok:
            raise APIException("Failed to get minimum accessible value from context")
        return buf.value

    def get_accessible_actions(self, context: JavaObject) -> AccessibleActions:
        actions = AccessibleActions()
        ok = self._wab.getAccessibleActions(self._vmID, context, byref(actions))
        if not ok:
            raise APIException("Failed to get accessible actions")
        return actions

    def do_accessible_actions(self, context: JavaObject, actions: AccessibleActionsToDo) -> None:
        index = c_int()
        ok = self._wab.doAccessibleActions(self._vmID, context, actions, byref(index))
        if not ok:
            raise APIException("Action failed at index={}".format(index))

    def set_text_contents(self, context: JavaObject, text: str) -> None:
        buf = create_unicode_buffer(text, MAX_STRING_SIZE)
        ok = self._wab.setTextContents(self._vmID, context, buf)
        if not ok:
            raise APIException("Failed to set field contents")

    def request_focus(self, context: JavaObject) -> None:
        ok = self._wab.requestFocus(self._vmID, context)
        if not ok:
            raise APIException("Failed to request focus")

    def get_accessible_key_bindings(self, context: JavaObject) -> AccessibleKeyBindings:
        key_bindings = AccessibleKeyBindings()
        ok = self._wab.getAccessibleKeyBindings(self._vmID, context, byref(key_bindings))
        if not ok:
            raise APIException("Failed to get key bindings")
        return key_bindings

    def is_same_object(self, context_from: JavaObject, context_to: JavaObject) -> bool:
        return self._wab.isSameObject(self._vmID, context_from, context_to)

    def get_virtual_accessible_name(self, context: JavaObject):
        buf = create_unicode_buffer(MAX_STRING_SIZE)
        ok = self._wab.getVirtualAccessibleName(self._vmID, context, buf, MAX_STRING_SIZE)
        if not ok:
            raise APIException("Failed to get virtual accessible name")
        return buf.value

    def register_callback(self, name: str, callback: Callable[[JavaObject], None]) -> None:
        logging.debug(f"Registering callback={name}")
        if name in self._context_callbacks:
            self._context_callbacks[name].append(callback)
        else:
            self._context_callbacks[name] = [callback]

    """
    Define the callback handlers
    """
    def _get_callback_func(self, name, wrapper, callback):
        def func(*args):
            callback(*args)
        runner = wrapper(func)
        setattr(self, name, runner)
        return runner

    def property_change(self, vmID: c_long, event: JavaObject, source: JavaObject, property, old_value, new_value):
        with _ReleaseEvent(self, vmID, "property_change", event, source):
            if 'property_change' in self._context_callbacks:
                for cp in self._context_callbacks['property_change']:
                    cp(source, property, old_value, new_value)

    def property_name_change(self, vmID: c_long, event: JavaObject, source: JavaObject, old_value: str, new_value: str):
        with _ReleaseEvent(self, vmID, "property_name_change", event, source):
            if 'property_name_change' in self._context_callbacks:
                for cp in self._context_callbacks['property_name_change']:
                    cp(source, old_value, new_value)

    def property_description_change(self, vmID: c_long, event: JavaObject, source: JavaObject, old_value: str, new_value: str):
        with _ReleaseEvent(self, vmID, "property_description_change", event, source):
            if 'property_description_change' in self._context_callbacks:
                for cp in self._context_callbacks['property_description_change']:
                    cp(source, old_value, new_value)

    def property_state_change(self, vmID: c_long, event: JavaObject, source: JavaObject, old_value: str, new_value: str):
        with _ReleaseEvent(self, vmID, "property_state_change", event, source):
            if 'property_state_change' in self._context_callbacks:
                for cp in self._context_callbacks['property_state_change']:
                    cp(source, old_value, new_value)

    def property_value_change(self, vmID: c_long, event: JavaObject, source: JavaObject, old_value: str, new_value: str):
        with _ReleaseEvent(self, vmID, "property_value_change", event, source):
            if 'property_value_change' in self._context_callbacks:
                for cp in self._context_callbacks['property_value_change']:
                    cp(source, old_value, new_value)

    def property_selection_change(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "property_selection_change", event, source):
            if 'property_selection_change' in self._context_callbacks:
                for cp in self._context_callbacks['property_selection_change']:
                    cp(source)

    def property_text_change(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "property_text_change", event, source):
            if 'property_text_change' in self._context_callbacks:
                for cp in self._context_callbacks['property_text_change']:
                    cp(source)

    def property_caret_change(self, vmID: c_long, event: JavaObject, source: JavaObject, old_pos: int, new_pos: int):
        with _ReleaseEvent(self, vmID, "set_property_caret_change", event, source):
            if 'set_property_caret_change' in self._context_callbacks:
                for cp in self._context_callbacks['set_property_caret_change']:
                    cp(source, old_pos, new_pos)

    def property_visible_data_change(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "property_visible_data_change", event, source):
            if 'property_visible_data_change' in self._context_callbacks:
                for cp in self._context_callbacks['property_visible_data_change']:
                    cp(source)

    def property_child_change(self, vmID: c_long, event: JavaObject, source: JavaObject, old_child: JavaObject, new_child: JavaObject):
        with _ReleaseEvent(self, vmID, "property_child_change", event, source):
            if 'property_child_change' in self._context_callbacks:
                for cp in self._context_callbacks['property_child_change']:
                    cp(source, old_child, new_child)

    def property_active_descendent_change(self, vmID: c_long, event: JavaObject, source: JavaObject, old_child: JavaObject, new_child: JavaObject):
        with _ReleaseEvent(self, vmID, "property_active_descendent_change", event, source):
            if 'property active descendent change' in self._context_callbacks:
                for cp in self._context_callbacks['property_active_descendent_change']:
                    cp(source, old_child, new_child)

    def property_table_model_change(self, vmID: c_long, event: JavaObject, source: JavaObject, old_value: str, new_value: str):
        with _ReleaseEvent(self, vmID, "property_table_model_change", event, source):
            if 'property_table_model_change' in self._context_callbacks:
                for cp in self._context_callbacks['property_table_model_change']:
                    cp(source, old_value, new_value)

    def menu_selected(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "menu_selected", event, source):
            if 'menu_selected' in self._context_callbacks:
                for cp in self._context_callbacks['menu_selected']:
                    cp(source)

    def menu_deselected(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "menu_deselected", event, source):
            if 'menu_deselected' in self._context_callbacks:
                for cp in self._context_callbacks['menu_deselected']:
                    cp(source)

    def menu_canceled(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "menu_canceled", event, source):
            if 'menu_canceled' in self._context_callbacks:
                for cp in self._context_callbacks['menu_canceled']:
                    cp(source)

    def focus_gained(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "focus_gained", event, source):
            if 'focus_gained' in self._context_callbacks:
                for cp in self._context_callbacks['focus_gained']:
                    cp(source)

    def focus_lost(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "focus_lost", event, source):
            if 'focus_lost' in self._context_callbacks:
                for cp in self._context_callbacks['focus_lost']:
                    cp(source)

    def caret_update(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "caret_update", event, source):
            if 'caret_update' in self._context_callbacks:
                for cp in self._context_callbacks['caret_update']:
                    cp(source)

    def mouse_clicked(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "mouse_clicked", event, source):
            if 'mouse_clicked' in self._context_callbacks:
                for cp in self._context_callbacks['mouse_clicked']:
                    cp(source)

    def mouse_entered(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "mouse_entered", event, source):
            if 'mouse_entered' in self._context_callbacks:
                for cp in self._context_callbacks['mouse_entered']:
                    cp(source)

    def mouse_exited(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "mouse_exited", event, source):
            if 'mouse_exited' in self._context_callbacks:
                for cp in self._context_callbacks['mouse_exited']:
                    cp(source)

    def mouse_pressed(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "mouse_pressed", event, source):
            if 'mouse_pressed' in self._context_callbacks:
                for cp in self._context_callbacks['mouse_pressed']:
                    cp(source)

    def mouse_released(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "mouse_released", event, source):
            if 'mouse_released' in self._context_callbacks:
                for cp in self._context_callbacks['mouse_released']:
                    cp(source)

    def popup_menu_canceled(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "popup_menu_canceled", event, source):
            if 'popup_menu_canceled' in self._context_callbacks:
                for cp in self._context_callbacks['popup_menu_canceled']:
                    cp(source)

    def popup_menu_will_become_invisible(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "popup_menu_will_become_invisible", event, source):
            if 'popup_menu_will_become_invisible' in self._context_callbacks:
                for cp in self._context_callbacks['popup_menu_will_become_invisible']:
                    cp(source)

    def popup_menu_will_become_visible(self, vmID: c_long, event: JavaObject, source: JavaObject):
        with _ReleaseEvent(self, vmID, "popup_menu_will_become_visible", event, source):
            if 'popup_menu_will_become_visible' in self._context_callbacks:
                for cp in self._context_callbacks['popup_menu_will_become_visible']:
                    cp(source)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['JABWrapper']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'java-access-bridge-wrapper',
    'version': '0.5.0',
    'description': 'Python wrapper for the Windows Java Access Bridge',
    'long_description': '\nPython wrapper around the Java Access Bridge / Windows Access Bridge.\n\n# Prerequisites\n\n* 64-bit Windows\n* Java >= 8 (https://docs.aws.amazon.com/corretto/latest/corretto-8-ug/downloads-list.html)\n* Python >= 3.7 (https://www.python.org/downloads/release/python-375/)\n* Install poetry: https://python-poetry.org/docs/\n\n# Test\n\nEnable the Java Access Bridge in windows\n\n* `C:\\path\\to\\java\\bin\\jabswitch -enable`.\n\nRun test script against simple Swing application\n\n* set environment variable `set RC_JAVA_ACCESS_BRIDGE_DLL=C:\\Program Files\\Java\\jre1.8.0_261\\bin\\WindowsAccessBridge-64.dll`\n* `poetry run python tests\\test.py`\n\n# Packaging\n\n* poetry build\n* poetry publish\n\n# TODO:\n\n* Support for 32-bit Java Access Bridge version\n* Add rest of the parsing functions\n* Better API to the ContextNode component\n\n',
    'author': 'Robocorp',
    'author_email': 'support@robocorp.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/robocorp/java-access-bridge-wrapper.git',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
